@@ -7,6 +7,7 @@
 #include <R.h>
 #include <Rdefines.h>
 #include <Rinternals.h>
+#include <Rmath.h>
 #include <R_ext/Rdynload.h>
 
 #include <dbarts/bartFit.hpp>
@@ -19,6 +20,7 @@
 
 #include <external/alloca.h>
 #include <external/linearAlgebra.h>
+#include <external/random.h>
 
 #include <set>
 #ifdef THREAD_SAFE_UNLOAD
@@ -27,12 +29,19 @@
 
 using std::size_t;
 
+extern "C" {
+  void R_init_dbarts(DllInfo* info);
+}
+
 namespace {
   using namespace dbarts;
   
   void initializeControlFromExpression(Control& control, SEXP controlExpr);
+  // SEXP createControlExpressionFromFit(const BARTFit& fit);
   void initializeModelFromExpression(Model& model, SEXP modelExpr, const Control& control);
+  // SEXP createModelExpressionFromFit(const BARTFit& fit);
   void initializeDataFromExpression(Data& data, SEXP dataExpr);
+  // SEXP createDataExpressionFromFit(const BARTFit& fit);
   
   void initializeStateFromExpression(const BARTFit& fit, State& state, SEXP stateExpr);
   SEXP createStateExpressionFromFit(const BARTFit& fit); // result has a protect count of 1
@@ -91,15 +100,22 @@ extern "C" {
 }
 
 namespace {
+  using namespace dbarts;
+  
   SEXP setResponse(SEXP fitExpr, SEXP y)
   {
     BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
     if (fit == NULL) error("dbarts_setY called on NULL external pointer.");
     
     if (!isReal(y)) error("y must be of type real.");
-    if ((size_t) length(y) != fit->data.numObservations) error("Length of new y does not match old.");
+    if (static_cast<size_t>(length(y)) != fit->data.numObservations) error("Length of new y does not match old.");
+    
+    if (fit->control.responseIsBinary) GetRNGstate();
+        
     fit->setResponse(REAL(y));
     
+    if (fit->control.responseIsBinary) PutRNGstate();
+    
     return NULL_USER_OBJECT;
   }
   
@@ -111,73 +127,192 @@ namespace {
     double* offset = NULL;
     if (isReal(offsetExpr)) {
       offset = REAL(offsetExpr);
-      if ((size_t) length(offsetExpr) != fit->data.numObservations) error("Length of new offset does not match y.");
+      if (static_cast<size_t>(length(offsetExpr)) != fit->data.numObservations) error("Length of new offset does not match y.");
     } else if (!isNull(offsetExpr) && !isS4Null(offsetExpr)) {
       error("offset must be of type real or NULL.");
     }
+    
+    if (fit->control.responseIsBinary) GetRNGstate();
+    
     fit->setOffset(offset);
     
+    if (fit->control.responseIsBinary) PutRNGstate();
+    
     return NULL_USER_OBJECT;
   }
   
-  SEXP setPredictor(SEXP fitExpr, SEXP x, SEXP jExpr)
+  SEXP setPredictor(SEXP fitExpr, SEXP x)
   {
     BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
     if (fit == NULL) error("dbarts_setPredictor called on NULL external pointer.");
     
-    if (!isReal(x)) error("x must be of type real.");
-    if ((size_t) length(x) != fit->data.numObservations) error("Length of new x does not match y.");
-    if (!isInteger(jExpr)) error("Column must be of type integer.");
-    if (length(jExpr) == 0) error("Length of column is 0.");
+    if (!isReal(x)) error("x must be of type real");
     
-    int j = INTEGER(jExpr)[0] - 1;
+    SEXP dimsExpr = GET_DIM(x);
     
-    if (j < 0 || j >= (int) fit->data.numPredictors) error("Column is out of range.");
+    if (isNull(dimsExpr) || LENGTH(dimsExpr) != 2) error("x must be a matrix, i.e. have two dimensions");
+    int* dims = INTEGER(dimsExpr);
     
-    fit->setPredictor(REAL(x), (size_t) j);
+    if (static_cast<size_t>(dims[0]) != fit->data.numObservations) error("number of rows in new x does not match y");
+    if (static_cast<size_t>(dims[1]) != fit->data.numPredictors) error("number of columns in new x does not match old");
     
-    return NULL_USER_OBJECT;
+    return ScalarLogical(fit->setPredictor(REAL(x)));
   }
   
-  SEXP setTestPredictor(SEXP fitExpr, SEXP x_test, SEXP jExpr)
+  SEXP updatePredictor(SEXP fitExpr, SEXP x, SEXP colsExpr)
+  {
+    BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
+    if (fit == NULL) error("dbarts_updatePredictor called on NULL external pointer");
+    
+    if (!isReal(x)) error("x must be of type real");
+    if (!isInteger(colsExpr)) error("columns must be of type integer");
+    
+    SEXP dimsExpr = GET_DIM(x);
+    int* dims = NULL;
+    
+    if (!isNull(dimsExpr)) {
+      int numDims = GET_LENGTH(dimsExpr);
+      
+      if (numDims != 1 && numDims != 2) error("x must be a vector or a matrix");
+      if (numDims == 2) dims = INTEGER(dimsExpr);
+    }
+    
+    if (length(colsExpr) == 0) error("length of columns is 0");
+
+    if (dims != NULL) {
+      if (static_cast<size_t>(dims[0]) != fit->data.numObservations) error("number of rows of new x does not match y");
+      if (dims[1] != length(colsExpr)) error("number of columns of new x does not match length of columns to replace");
+    } else {
+      if (static_cast<size_t>(length(x)) != fit->data.numObservations) error("length of new x does not match y");
+    }
+    
+    
+    int* colsInt = INTEGER(colsExpr);
+    size_t numCols = static_cast<size_t>(LENGTH(colsExpr));
+    size_t* cols = ext_stackAllocate(numCols, size_t);
+    for (size_t i = 0 ; i < numCols; ++i) {
+      cols[i] = static_cast<size_t>(colsInt[i] - 1);
+      if (static_cast<size_t>(cols[i]) >= fit->data.numPredictors) error("column '%d' is out of range", colsInt[i]);
+    }
+    
+    bool result = fit->updatePredictors(REAL(x), cols, numCols);
+    
+    ext_stackFree(cols);
+    
+    return ScalarLogical(result);
+  }
+  
+  SEXP setTestPredictor(SEXP fitExpr, SEXP x_test)
   {
     BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
     if (fit == NULL) error("dbarts_setTestPredictor called on NULL external pointer.");
     
-    if (fit->data.X_test == NULL) error("Test matrix must exist at object creation to be updated.");
-    
-    if (!isReal(x_test)) error("x must be of type real.");
-    if ((size_t) length(x_test) != fit->data.numTestObservations) error("Length of new x_test does not match old.");
-    if (!isInteger(jExpr)) error("Column must be of type integer.");
-    if (length(jExpr) == 0) error("Length of column is 0.");
-    
-    int j = INTEGER(jExpr)[0] - 1;
-    
-    if (j < 0 || j >= (int) fit->data.numPredictors) error("Column is out of range.");
-    
-    fit->setTestPredictor(REAL(x_test), (size_t) j);
-    
-    return NULL_USER_OBJECT;
-  }
-  
-  SEXP setTestPredictors(SEXP fitExpr, SEXP x_test)
-  {
-    BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
-    if (fit == NULL) error("dbarts_setTestPredictors called on NULL external pointer.");
-    
     if (isNull(x_test) || isS4Null(x_test)) {
-      fit->setTestPredictors(NULL, 0);
+      fit->setTestPredictor(NULL, 0);
     
       return NULL_USER_OBJECT;
     }
     
-    if (!isReal(x_test)) error("x.test must be of type real.");
+    if (!isReal(x_test)) error("x.test must be of type real");
     SEXP dimsExpr = GET_DIM(x_test);
-    if (GET_LENGTH(dimsExpr) != 2) error("x.test must be a matrix, i.e. have two dimensions.");
+    if (GET_LENGTH(dimsExpr) != 2) error("x.test must be a matrix, i.e. have two dimensions");
     int* dims = INTEGER(dimsExpr);
-    if ((size_t) dims[1] != fit->data.numPredictors) error("Number of columns of x.test and x must be equal.");
+    if (static_cast<size_t>(dims[1]) != fit->data.numPredictors) error("number of columns of x.test and x must be equal");
     
-    fit->setTestPredictors(REAL(x_test), (size_t) dims[0]);
+    fit->setTestPredictor(REAL(x_test), static_cast<size_t>(dims[0]));
+    
+    return NULL_USER_OBJECT;
+  }
+  
+  SEXP setTestOffset(SEXP fitExpr, SEXP offset_test)
+  {
+    BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
+    if (fit == NULL) error("dbarts_setTestOffset called on NULL external pointer.");
+    
+    if (isNull(offset_test)) {
+      fit->setTestOffset(NULL);
+    } else {
+      if (!isReal(offset_test)) error("offset.test must be of type real");
+      if (fit->data.numTestObservations != static_cast<size_t>(GET_LENGTH(offset_test))) error("length of offset.test must equal number of rows in x.test");
+      fit->setTestOffset(REAL(offset_test));
+    }
+    
+    return NULL_USER_OBJECT;
+  }
+  
+  SEXP setTestPredictorAndOffset(SEXP fitExpr, SEXP x_test, SEXP offset_test)
+  {
+    BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
+    if (fit == NULL) error("dbarts_setTestPredictorAndOffset called on NULL external pointer.");
+    
+    if (isNull(x_test) || isS4Null(x_test)) {
+      fit->setTestPredictor(NULL, 0);
+    
+      return NULL_USER_OBJECT;
+    }
+    
+    if (!isReal(x_test)) error("x.test must be of type real");
+    SEXP dimsExpr = GET_DIM(x_test);
+    if (GET_LENGTH(dimsExpr) != 2) error("x.test must be a matrix, i.e. have two dimensions");
+    int* dims = INTEGER(dimsExpr);
+    if (static_cast<size_t>(dims[1]) != fit->data.numPredictors) error("number of columns of x.test and x must be equal");
+    
+    if (isNull(offset_test)) {
+      fit->setTestPredictorAndOffset(REAL(x_test), NULL, static_cast<size_t>(dims[0]));
+    } else {
+      if (!isReal(offset_test)) error("offset.test must be of type real");
+      if (GET_LENGTH(offset_test) == 1 && ISNA(REAL(offset_test)[0])) {
+        fit->setTestPredictor(REAL(x_test), static_cast<size_t>(dims[0]));
+      } else {
+        if (GET_LENGTH(offset_test) != dims[0]) error("length of offset.test must equal number of rows in x.test");
+        fit->setTestPredictorAndOffset(REAL(x_test), REAL(offset_test), static_cast<size_t>(dims[0]));
+      }
+    }
+    
+    return NULL_USER_OBJECT;
+  }
+  
+  SEXP updateTestPredictor(SEXP fitExpr, SEXP x_test, SEXP colsExpr)
+  {
+    BARTFit* fit = static_cast<BARTFit*>(R_ExternalPtrAddr(fitExpr));
+    if (fit == NULL) error("dbarts_updateTestPredictor called on NULL external pointer.");
+    
+    if (fit->data.X_test == NULL) error("test matrix must exist at object creation to be updated");
+    
+    if (!isReal(x_test)) error("x must be of type real");
+    if (!isInteger(colsExpr)) error("columns must be of type integer");
+    
+    SEXP dimsExpr = GET_DIM(x_test);
+    int* dims = NULL;
+    
+    if (!isNull(dimsExpr)) {
+      int numDims = GET_LENGTH(dimsExpr);
+      
+      if (numDims != 1 && numDims != 2) error("x must be a vector or a matrix");
+      if (numDims == 2) dims = INTEGER(dimsExpr);
+    }
+    
+    if (length(colsExpr) == 0) error("length of columns is 0");
+
+    if (dims != NULL) {
+      if (static_cast<size_t>(dims[0]) != fit->data.numTestObservations) error("number of rows of new x does not match old x.test");
+      if (dims[1] != length(colsExpr)) error("number of columns of new x does not match length of columns to replace");
+    } else {
+      if (static_cast<size_t>(length(x_test)) != fit->data.numTestObservations) error("length of new x does not match old x.test");
+    }
+    
+    
+    int* colsInt = INTEGER(colsExpr);
+    size_t numCols = static_cast<size_t>(LENGTH(colsExpr));
+    size_t* cols = ext_stackAllocate(numCols, size_t);
+    for (size_t i = 0 ; i < numCols; ++i) {
+      cols[i] = static_cast<size_t>(colsInt[i] - 1);
+      if (cols[i] >= fit->data.numPredictors) error("column '%d' is out of range", colsInt[i]);
+    }
+    
+    fit->updateTestPredictors(REAL(x_test), cols, numCols);
+    
+    ext_stackFree(cols);
     
     return NULL_USER_OBJECT;
   }
@@ -237,10 +372,11 @@ namespace {
     BARTFit* fit = new BARTFit(control, model, data);
     
     SEXP result = PROTECT(R_MakeExternalPtr(fit, NULL_USER_OBJECT, NULL_USER_OBJECT));
-    R_RegisterCFinalizerEx(result, fitFinalizer, (Rboolean) FALSE);
+    R_RegisterCFinalizerEx(result, fitFinalizer, static_cast<Rboolean>(FALSE));
     
 #ifdef THREAD_SAFE_UNLOAD
     pthread_mutex_lock(&fitMutex);
+    Rprintf("creating   %p\n", fit);
 #endif
     activeFits->insert(result);
 #ifdef THREAD_SAFE_UNLOAD
@@ -296,13 +432,13 @@ namespace {
     if (length(numBurnInExpr) == 0) error("Number of burn-in steps must be of length at least 1.");
     i_temp = INTEGER(numBurnInExpr)[0];
     if (i_temp != NA_INTEGER && i_temp < 0) error("Number of burn-in steps must be non-negative.");
-    numBurnIn = i_temp == NA_INTEGER ? fit->control.numBurnIn : (size_t) i_temp;
+    numBurnIn = i_temp == NA_INTEGER ? fit->control.numBurnIn : static_cast<size_t>(i_temp);
     
     if (!isInteger(numSamplesExpr)) error("Number of samples must be of integer type.");
     if (length(numSamplesExpr) == 0) error("Number of samples must be of length at least 1.");
     i_temp = INTEGER(numSamplesExpr)[0];
     if (i_temp != NA_INTEGER && i_temp <= 0) error("Number of samples must be positive.");
-    numSamples = i_temp == NA_INTEGER ? fit->control.numSamples : (size_t) i_temp;
+    numSamples = i_temp == NA_INTEGER ? fit->control.numSamples : static_cast<size_t>(i_temp);
     
     GetRNGstate();
         
@@ -318,46 +454,46 @@ namespace {
     int protectCount = 3;
     
     SEXP resultExpr = PROTECT(allocVector(VECSXP, 4));
-    SET_VECTOR_ELT(resultExpr, 0, allocVector(REALSXP, (int) bartResults->getNumSigmaSamples()));
-    SET_VECTOR_ELT(resultExpr, 1, allocVector(REALSXP, (int) bartResults->getNumTrainingSamples()));
+    SET_VECTOR_ELT(resultExpr, 0, allocVector(REALSXP, static_cast<int>(bartResults->getNumSigmaSamples())));
+    SET_VECTOR_ELT(resultExpr, 1, allocVector(REALSXP, static_cast<int>(bartResults->getNumTrainingSamples())));
     if (fit->data.numTestObservations > 0)
-      SET_VECTOR_ELT(resultExpr, 2, allocVector(REALSXP, (int) bartResults->getNumTestSamples()));
+      SET_VECTOR_ELT(resultExpr, 2, allocVector(REALSXP, static_cast<int>(bartResults->getNumTestSamples())));
     else
       SET_VECTOR_ELT(resultExpr, 2, NULL_USER_OBJECT);
-    SET_VECTOR_ELT(resultExpr, 3, allocVector(INTSXP, (int) bartResults->getNumVariableCountSamples()));
+    SET_VECTOR_ELT(resultExpr, 3, allocVector(INTSXP, static_cast<int>(bartResults->getNumVariableCountSamples())));
     
     SEXP sigmaSamples = VECTOR_ELT(resultExpr, 0);
-    std::memcpy(REAL(sigmaSamples), (const double*) bartResults->sigmaSamples, bartResults->getNumSigmaSamples() * sizeof(double));
+    std::memcpy(REAL(sigmaSamples), const_cast<const double*>(bartResults->sigmaSamples), bartResults->getNumSigmaSamples() * sizeof(double));
     
     SEXP trainingSamples = VECTOR_ELT(resultExpr, 1);
     dimsExpr = PROTECT(dimsExpr = allocVector(INTSXP, 2));
     dims = INTEGER(dimsExpr);
-    dims[0] = (int) bartResults->numObservations;
-    dims[1] = (int) bartResults->numSamples;
+    dims[0] = static_cast<int>(bartResults->numObservations);
+    dims[1] = static_cast<int>(bartResults->numSamples);
     setAttrib(trainingSamples, R_DimSymbol, dimsExpr);
-    std::memcpy(REAL(trainingSamples), (const double*) bartResults->trainingSamples, bartResults->getNumTrainingSamples() * sizeof(double));
+    std::memcpy(REAL(trainingSamples), const_cast<const double*>(bartResults->trainingSamples), bartResults->getNumTrainingSamples() * sizeof(double));
     
     if (fit->data.numTestObservations > 0) {
       SEXP testSamples = VECTOR_ELT(resultExpr, 2);
       dimsExpr = PROTECT(dimsExpr = allocVector(INTSXP, 2));
       dims = INTEGER(dimsExpr);
-      dims[0] = (int) bartResults->numTestObservations;
-      dims[1] = (int) bartResults->numSamples;
+      dims[0] = static_cast<int>(bartResults->numTestObservations);
+      dims[1] = static_cast<int>(bartResults->numSamples);
       setAttrib(testSamples, R_DimSymbol, dimsExpr);
-      std::memcpy(REAL(testSamples), (const double*) bartResults->testSamples, bartResults->getNumTestSamples() * sizeof(double));
+      std::memcpy(REAL(testSamples), const_cast<const double*>(bartResults->testSamples), bartResults->getNumTestSamples() * sizeof(double));
       ++protectCount;
     }
     
     SEXP variableCountSamples = VECTOR_ELT(resultExpr, 3);
     dimsExpr = PROTECT(dimsExpr = allocVector(INTSXP, 2));
     dims = INTEGER(dimsExpr);
-    dims[0] = (int) bartResults->numPredictors;
-    dims[1] = (int) bartResults->numSamples;
+    dims[0] = static_cast<int>(bartResults->numPredictors);
+    dims[1] = static_cast<int>(bartResults->numSamples);
     setAttrib(variableCountSamples, R_DimSymbol, dimsExpr);
     int* variableCountStorage = INTEGER(variableCountSamples);
     size_t length = bartResults->getNumVariableCountSamples();
     // these likely need to be down-sized from 64 to 32 bits
-    for (size_t i = 0; i < length; ++i) variableCountStorage[i] = (int) bartResults->variableCountSamples[i];
+    for (size_t i = 0; i < length; ++i) variableCountStorage[i] = static_cast<int>(bartResults->variableCountSamples[i]);
     
     setAttrib(resultExpr, R_NamesSymbol, namesExpr = allocVector(STRSXP, 4));
     SET_STRING_ELT(namesExpr, 0, mkChar("sigma"));
@@ -399,6 +535,11 @@ namespace {
     return NULL_USER_OBJECT;
   }
   
+  SEXP deepCopy(SEXP obj)
+  {
+    return duplicate(obj);
+  }
+  
   // as of R 3.1, auto-unload never gets called so screw that
   
 /*  void R_unload_dbarts(DllInfo* info)
@@ -414,75 +555,91 @@ namespace {
     pthread_mutex_unlock(&fitMutex);
     pthread_mutex_destroy(&fitMutex);
   }*/
+  
+#define DEF_FUNC(_N_, _F_, _A_) { _N_, reinterpret_cast<DL_FUNC>(&_F_), _A_ }
 
   R_CallMethodDef R_callMethods[] = {
-    { "dbarts_create", (DL_FUNC) &create, 3 },
-    { "dbarts_run", (DL_FUNC) &run, 3 },
-    { "dbarts_setResponse", (DL_FUNC) &setResponse, 2 },
-    { "dbarts_setOffset", (DL_FUNC) &setOffset, 2 },
-    { "dbarts_setPredictor", (DL_FUNC) &setPredictor, 3 },
-    { "dbarts_setTestPredictor", (DL_FUNC) &setTestPredictor, 3 },
-    { "dbarts_setTestPredictors", (DL_FUNC) &setTestPredictors, 2 },
-    { "dbarts_setControl", (DL_FUNC) &setControl, 2 },
-    { "dbarts_isValidPointer", (DL_FUNC) &isValidPointer, 1 },
-    { "dbarts_createState", (DL_FUNC) &createState, 1 },
-    { "dbarts_storeState", (DL_FUNC) &storeState, 2 },
-    { "dbarts_restoreState", (DL_FUNC) &restoreState, 2},
-    { "dbarts_finalize", (DL_FUNC) &finalize, 0 },
+    DEF_FUNC("dbarts_create", create, 3),
+    DEF_FUNC("dbarts_run", run, 3),
+    DEF_FUNC("dbarts_setResponse", setResponse, 2),
+    DEF_FUNC("dbarts_setOffset", setOffset, 2),
+    DEF_FUNC("dbarts_setPredictor", setPredictor, 2),
+    DEF_FUNC("dbarts_updatePredictor", updatePredictor, 3),
+    DEF_FUNC("dbarts_setTestPredictor", setTestPredictor, 2),
+    DEF_FUNC("dbarts_setTestOffset", setTestOffset, 2),
+    DEF_FUNC("dbarts_setTestPredictorAndOffset", setTestPredictorAndOffset, 3),
+    DEF_FUNC("dbarts_updateTestPredictor", updateTestPredictor, 3),
+    DEF_FUNC("dbarts_setControl", setControl, 2),
+    DEF_FUNC("dbarts_isValidPointer", isValidPointer, 1),
+    DEF_FUNC("dbarts_createState", createState, 1),
+    DEF_FUNC("dbarts_storeState", storeState, 2),
+    DEF_FUNC("dbarts_restoreState", restoreState, 2),
+    DEF_FUNC("dbarts_finalize", finalize, 0),
+    DEF_FUNC("dbarts_deepCopy", deepCopy, 1),
     { NULL, NULL, 0 }
   };
+
+#undef DEF_FUNC
   
   struct C_CallMethodDef {
-    const char* package;
     const char* name;
     DL_FUNC function;
   };
   
+#define DEF_FUNC(_N_, _F_) { _N_, reinterpret_cast<DL_FUNC>(&_F_) }
+  
   C_CallMethodDef C_callMethods[] = {
-    { "dbarts", "createCGMPrior", (DL_FUNC) dbarts_createCGMPrior },
-    { "dbarts", "createCGMPriorFromOptions", (DL_FUNC) dbarts_createCGMPriorFromOptions },
-    { "dbarts", "destroyCGMPrior", (DL_FUNC) dbarts_destroyCGMPrior },
-    { "dbarts", "initializeCGMPriorFromOptions", (DL_FUNC) dbarts_initializeCGMPriorFromOptions },
-    { "dbarts", "invalidateCGMPrior", (DL_FUNC) dbarts_invalidateCGMPrior },
+    DEF_FUNC("createCGMPrior", dbarts_createCGMPrior),
+    DEF_FUNC("createCGMPriorFromOptions", dbarts_createCGMPriorFromOptions),
+    DEF_FUNC("destroyCGMPrior", dbarts_destroyCGMPrior),
+    DEF_FUNC("initializeCGMPriorFromOptions", dbarts_initializeCGMPriorFromOptions),
+    DEF_FUNC("invalidateCGMPrior", dbarts_invalidateCGMPrior),
     
-    { "dbarts", "createNormalPrior", (DL_FUNC) dbarts_createNormalPrior },
-    { "dbarts", "createNormalPriorFromOptions", (DL_FUNC) dbarts_createNormalPriorFromOptions },
-    { "dbarts", "destroyNormalPrior", (DL_FUNC) dbarts_destroyNormalPrior },
-    { "dbarts", "initializeNormalPriorFromOptions", (DL_FUNC) dbarts_initializeNormalPriorFromOptions },
-    { "dbarts", "invalidateNormalPrior", (DL_FUNC) dbarts_invalidateNormalPrior },
+    DEF_FUNC("createNormalPrior", dbarts_createNormalPrior),
+    DEF_FUNC("createNormalPriorFromOptions", dbarts_createNormalPriorFromOptions),
+    DEF_FUNC("destroyNormalPrior", dbarts_destroyNormalPrior),
+    DEF_FUNC("initializeNormalPriorFromOptions", dbarts_initializeNormalPriorFromOptions),
+    DEF_FUNC("invalidateNormalPrior", dbarts_invalidateNormalPrior),
     
-    { "dbarts", "createChiSquaredPrior", (DL_FUNC) dbarts_createChiSquaredPrior },
-    { "dbarts", "createChiSquaredPriorFromOptions", (DL_FUNC) dbarts_createChiSquaredPriorFromOptions },
-    { "dbarts", "destroyChiSquaredPrior", (DL_FUNC) dbarts_destroyChiSquaredPrior },
-    { "dbarts", "initializeChiSquaredPriorFromOptions", (DL_FUNC) dbarts_initializeChiSquaredPriorFromOptions },
-    { "dbarts", "invalidateChiSquaredPrior", (DL_FUNC) dbarts_invalidateChiSquaredPrior },
+    DEF_FUNC("createChiSquaredPrior", dbarts_createChiSquaredPrior),
+    DEF_FUNC("createChiSquaredPriorFromOptions", dbarts_createChiSquaredPriorFromOptions),
+    DEF_FUNC("destroyChiSquaredPrior", dbarts_destroyChiSquaredPrior),
+    DEF_FUNC("initializeChiSquaredPriorFromOptions", dbarts_initializeChiSquaredPriorFromOptions),
+    DEF_FUNC("invalidateChiSquaredPrior", dbarts_invalidateChiSquaredPrior),
 
-    { "dbarts", "createFit", (DL_FUNC) dbarts_createFit },
-    { "dbarts", "initializeFit", (DL_FUNC) dbarts_initializeFit },
-    { "dbarts", "destroyFit", (DL_FUNC) dbarts_destroyFit },
-    { "dbarts", "invalidateFit", (DL_FUNC) dbarts_invalidateFit },
+    DEF_FUNC("createFit", dbarts_createFit),
+    DEF_FUNC("initializeFit", dbarts_initializeFit),
+    DEF_FUNC("destroyFit", dbarts_destroyFit),
+    DEF_FUNC("invalidateFit", dbarts_invalidateFit),
     
-    { "dbarts", "runSampler", (DL_FUNC) dbarts_runSampler },
-    { "dbarts", "runSamplerForIterations", (DL_FUNC) dbarts_runSamplerForIterations },
-    { "dbarts", "setResponse", (DL_FUNC) dbarts_setResponse },
-    { "dbarts", "setOffset", (DL_FUNC) dbarts_setOffset },
-    { "dbarts", "setPredictor", (DL_FUNC) dbarts_setPredictor },
-    { "dbarts", "setTestPredictor", (DL_FUNC) dbarts_setTestPredictor },
-    { "dbarts", "setTestPredictors", (DL_FUNC) dbarts_setTestPredictors },
-    { NULL, NULL, 0 }
+    DEF_FUNC("runSampler", dbarts_runSampler),
+    DEF_FUNC("runSamplerForIterations", dbarts_runSamplerForIterations),
+    DEF_FUNC("setResponse", dbarts_setResponse),
+    DEF_FUNC("setOffset", dbarts_setOffset),
+    DEF_FUNC("setPredictor", dbarts_setPredictor),
+    DEF_FUNC("updatePredictor", dbarts_updatePredictor),
+    DEF_FUNC("updatePredictors", dbarts_updatePredictors),
+    DEF_FUNC("setTestPredictor", dbarts_setTestPredictor),
+    DEF_FUNC("setTestOffset", dbarts_setTestOffset),
+    DEF_FUNC("setTestPredictorsAndOffset", dbarts_setTestPredictorAndOffset),
+    DEF_FUNC("updateTestPredictor", dbarts_updateTestPredictor),
+    DEF_FUNC("updateTestPredictors", dbarts_updateTestPredictors),
+    { NULL, 0 }
   };
   
+#undef DEF_FUNC
+  
 } // end anonymous namespace
 
 extern "C" {
   void R_init_dbarts(DllInfo* info)
   {
     R_registerRoutines(info, NULL, R_callMethods, NULL, NULL);
-    R_useDynamicSymbols(info, FALSE);
+    R_useDynamicSymbols(info, static_cast<Rboolean>(FALSE));
     
     C_CallMethodDef* method = C_callMethods;
-    while (method->package != NULL) {
-      R_RegisterCCallable(method->package, method->name, method->function);
+    while (method->name != NULL) {
+      R_RegisterCCallable("dbarts", method->name, method->function);
       ++method;
     }
     
@@ -508,7 +665,7 @@ namespace {
   SEXP SET_DIMS(SEXP obj, int numRows, int numCols)
   {
     SEXP dimsExp = NEW_INTEGER(2);
-    int *dims = INTEGER(dimsExp);
+    int* dims = INTEGER(dimsExp);
     dims[0] = numRows;
     dims[1] = numCols;
     
@@ -568,7 +725,7 @@ namespace {
     i_temp = INTEGER(slotExpr)[0];
     if (i_temp == NA_INTEGER) error("Number of samples cannot be NA.");
     if (i_temp <= 0) error("Number of samples must be positive.");
-    control.numSamples = (size_t) i_temp;
+    control.numSamples = static_cast<size_t>(i_temp);
     
     slotExpr = GET_ATTR(controlExpr, install("n.burn"));
     if (!isInteger(slotExpr)) error("Number of burn-in steps must be of integer type.");
@@ -576,7 +733,7 @@ namespace {
     i_temp = INTEGER(slotExpr)[0];
     if (i_temp == NA_INTEGER) i_temp = 0;
     if (i_temp < 0) error("Number of burn-in steps must be non-negative.");
-    control.numBurnIn = (size_t) i_temp;
+    control.numBurnIn = static_cast<size_t>(i_temp);
     
     slotExpr = GET_ATTR(controlExpr, install("n.trees"));
     if (!isInteger(slotExpr)) error("Number of trees must be of integer type.");
@@ -584,7 +741,7 @@ namespace {
     i_temp = INTEGER(slotExpr)[0];
     if (i_temp == NA_INTEGER) error("Number of trees cannot be NA.");
     if (i_temp <= 0) error("Number of trees must be positive.");
-    control.numTrees = (size_t) i_temp;
+    control.numTrees = static_cast<size_t>(i_temp);
     
     slotExpr = GET_ATTR(controlExpr, install("n.threads"));
     if (!isInteger(slotExpr)) error("Number of threads must be of integer type.");
@@ -592,7 +749,7 @@ namespace {
     i_temp = INTEGER(slotExpr)[0];
     if (i_temp == NA_INTEGER) i_temp = 1;
     if (i_temp <= 0) error("Number of threads must be positive.");
-    control.numThreads = (size_t) i_temp;
+    control.numThreads = static_cast<size_t>(i_temp);
     
     slotExpr = GET_ATTR(controlExpr, install("n.thin"));
     if (!isInteger(slotExpr)) error("Tree thinning rate must be of integer type.");
@@ -600,7 +757,7 @@ namespace {
     i_temp = INTEGER(slotExpr)[0];
     if (i_temp == NA_INTEGER) i_temp = 1;
     if (i_temp < 0) error("Tree thinning rate must be non-negative.");
-    control.treeThinningRate = (uint32_t) i_temp;
+    control.treeThinningRate = static_cast<uint32_t>(i_temp);
     
     
     slotExpr = GET_ATTR(controlExpr, install("printEvery"));
@@ -609,7 +766,7 @@ namespace {
     i_temp = INTEGER(slotExpr)[0];
     if (i_temp != NA_INTEGER) {
       if (i_temp <= 0) error("Print every must be positive.");
-      control.printEvery = (uint32_t) i_temp;
+      control.printEvery = static_cast<uint32_t>(i_temp);
     }
     
     slotExpr = GET_ATTR(controlExpr, install("printCutoffs"));
@@ -618,7 +775,19 @@ namespace {
     i_temp = INTEGER(slotExpr)[0];
     if (i_temp == NA_INTEGER) i_temp = 0;
     if (i_temp < 0) error("Print cutoffs must be non-negative.");
-    control.printCutoffs = (uint32_t) i_temp;
+    control.printCutoffs = static_cast<uint32_t>(i_temp);
+    
+    if (control.rng == NULL) {
+      ext_rng_userFunction uniformFunction;
+      uniformFunction.f.stateless = &unif_rand;
+      uniformFunction.state = NULL;
+      control.rng = ext_rng_create(EXT_RNG_ALGORITHM_USER_UNIFORM, &uniformFunction);
+      
+      ext_rng_userFunction normalFunction;
+      normalFunction.f.stateless = &norm_rand;
+      normalFunction.state = NULL;
+      ext_rng_setStandardNormalAlgorithm(control.rng, EXT_RNG_STANDARD_NORMAL_USER_NORM, &normalFunction);
+    }
   }
   
   void initializeModelFromExpression(Model& model, SEXP modelExpr, const Control& control)
@@ -722,19 +891,19 @@ namespace {
     if (!isReal(slotExpr)) error("y must be of type real.");
     if (length(slotExpr) == 0) error("Length of y must be greater than 0.");
     data.y = REAL(slotExpr);
-    data.numObservations = (size_t) length(slotExpr);
+    data.numObservations = static_cast<size_t>(length(slotExpr));
     
     slotExpr = GET_ATTR(dataExpr, install("x"));
     if (!isReal(slotExpr)) error("x must be of type real.");
     dims = INTEGER(GET_ATTR(slotExpr, R_DimSymbol));
     if (dims == NULL || length(GET_ATTR(slotExpr, R_DimSymbol)) != 2) error("x must be a matrix, i.e. have two dimensions.");
-    if ((size_t) dims[0] != data.numObservations) error("Number of rows of x and length of y must be equal.");
+    if (static_cast<size_t>(dims[0]) != data.numObservations) error("Number of rows of x and length of y must be equal.");
     data.X = REAL(slotExpr);
-    data.numPredictors = (size_t) dims[1];
+    data.numPredictors = static_cast<size_t>(dims[1]);
     
     slotExpr = GET_ATTR(dataExpr, install("varTypes"));
     if (!isInteger(slotExpr)) error("Variable types must be of type integer.");
-    if ((size_t) length(slotExpr) != data.numPredictors) error("Length of variable types must equal number of columns in x.");
+    if (static_cast<size_t>(length(slotExpr)) != data.numPredictors) error("Length of variable types must equal number of columns in x.");
     int* i_variableTypes = INTEGER(slotExpr);
     VariableType* variableTypes = new VariableType[data.numPredictors];
     for (size_t i = 0; i < data.numPredictors; ++i) variableTypes[i] = (i_variableTypes[i] == 0 ? ORDINAL : CATEGORICAL);
@@ -748,9 +917,9 @@ namespace {
       if (!isReal(slotExpr)) error ("x.test must be of type real.");
       dims = INTEGER(GET_ATTR(slotExpr, R_DimSymbol));
       if (dims == NULL || length(GET_ATTR(slotExpr, R_DimSymbol)) != 2) error("x.test must be a matrix, i.e. have two dimensions.");
-      if ((size_t) dims[1] != data.numPredictors) error("Number of columns of x.test and x must be equal.");
+      if (static_cast<size_t>(dims[1]) != data.numPredictors) error("Number of columns of x.test and x must be equal.");
       data.X_test = REAL(slotExpr);
-      data.numTestObservations = (size_t) dims[0];
+      data.numTestObservations = static_cast<size_t>(dims[0]);
     }
     
     slotExpr = GET_ATTR(dataExpr, install("weights"));
@@ -758,7 +927,7 @@ namespace {
       data.weights = NULL;
     } else {
       if (!isReal(slotExpr)) error("weights must be of type real.");
-      if ((size_t) length(slotExpr) != data.numObservations) error("Length of weights must equal length of y.");
+      if (static_cast<size_t>(length(slotExpr)) != data.numObservations) error("Length of weights must equal length of y.");
       data.weights = REAL(slotExpr);
     }
     
@@ -767,10 +936,19 @@ namespace {
       data.offset = NULL;
     } else {
       if (!isReal(slotExpr)) error("offset must be of type real.");
-      if ((size_t) length(slotExpr) != data.numObservations) error("Length of offset must equal length of y.");
+      if (static_cast<size_t>(length(slotExpr)) != data.numObservations) error("Length of offset must equal length of y.");
       data.offset = REAL(slotExpr);
     }
     
+    slotExpr = GET_ATTR(dataExpr, install("offset.test"));
+    if (isS4Null(slotExpr) || isNull(slotExpr) || length(slotExpr) == 0) {
+      data.testOffset = NULL;
+    } else {
+      if (!isReal(slotExpr)) error("Test offset must be of type real.");
+      if (static_cast<size_t>(length(slotExpr)) != data.numTestObservations) error("Length of test offset must equal number of test rows.");
+      data.testOffset = REAL(slotExpr);
+    }
+    
     slotExpr = GET_ATTR(dataExpr, install("sigma"));
     if (!isReal(slotExpr)) error("sigma estimate must be of type real.");
     if (length(slotExpr) != 1) error("sigma estimate must be of length 1.");
@@ -782,10 +960,10 @@ namespace {
     
     slotExpr = GET_ATTR(dataExpr, install("n.cuts"));
     if (!isInteger(slotExpr)) error("Maximum number of cuts must be of integer type.");
-    if ((size_t) length(slotExpr) != data.numPredictors) error("Length of maximum number of cuts and the number of columns of x must be equal.");
+    if (static_cast<size_t>(length(slotExpr)) != data.numPredictors) error("Length of maximum number of cuts and the number of columns of x must be equal.");
     int* i_maxNumCuts = INTEGER(slotExpr);
     uint32_t* maxNumCuts = new uint32_t[data.numPredictors];
-    for (size_t i = 0; i < data.numPredictors; ++i) maxNumCuts[i] = (uint32_t) i_maxNumCuts[i];
+    for (size_t i = 0; i < data.numPredictors; ++i) maxNumCuts[i] = static_cast<uint32_t>(i_maxNumCuts[i]);
     data.maxNumCuts = maxNumCuts;
   }
   
@@ -797,17 +975,17 @@ namespace {
     
     SEXP result = PROTECT(NEW_OBJECT(MAKE_CLASS("dbartsState")));
     
-    SEXP slotExpr = ALLOC_SLOT(result, install("fit.tree"), REALSXP, (int) (data.numObservations * control.numTrees));
-    SET_DIMS(slotExpr, (int) data.numObservations, (int) control.numTrees);
+    SEXP slotExpr = ALLOC_SLOT(result, install("fit.tree"), REALSXP, static_cast<int>(data.numObservations * control.numTrees));
+    SET_DIMS(slotExpr, static_cast<int>(data.numObservations), static_cast<int>(control.numTrees));
     std::memcpy(REAL(slotExpr), state.treeFits, data.numObservations * control.numTrees * sizeof(double));
     
-    slotExpr = ALLOC_SLOT(result, install("fit.total"), REALSXP, (int) data.numObservations);
+    slotExpr = ALLOC_SLOT(result, install("fit.total"), REALSXP, static_cast<int>(data.numObservations));
     std::memcpy(REAL(slotExpr), state.totalFits, data.numObservations * sizeof(double));
     
     if (data.numTestObservations == 0) {
       SET_SLOT(result, install("fit.test"), NULL_USER_OBJECT);
     } else {
-      slotExpr = ALLOC_SLOT(result, install("fit.test"), REALSXP, (int) data.numTestObservations);
+      slotExpr = ALLOC_SLOT(result, install("fit.test"), REALSXP, static_cast<int>(data.numTestObservations));
       std::memcpy(REAL(slotExpr), state.totalTestFits, data.numTestObservations * sizeof(double));
     }
     
@@ -817,11 +995,11 @@ namespace {
     slotExpr = ALLOC_SLOT(result, install("runningTime"), REALSXP, 1);
     REAL(slotExpr)[0] = state.runningTime;
     
-    slotExpr = ALLOC_SLOT(result, install("trees"), STRSXP, (int) control.numTrees);
+    slotExpr = ALLOC_SLOT(result, install("trees"), STRSXP, static_cast<int>(control.numTrees));
 
     const char** treeStrings = const_cast<const char**>(state.createTreeStrings(fit));
     for (size_t i = 0; i < control.numTrees; ++i) {
-      SET_STRING_ELT(slotExpr, (int) i, CREATE_STRING_VECTOR(treeStrings[i]));
+      SET_STRING_ELT(slotExpr, static_cast<int>(i), CREATE_STRING_VECTOR(treeStrings[i]));
       delete [] treeStrings[i];
     }
     delete [] treeStrings;
@@ -839,16 +1017,16 @@ namespace {
     SEXP dimsExpr = GET_DIM(slotExpr);
     if (GET_LENGTH(dimsExpr) != 2) error("Dimensions of state@fit.tree indicate that it is not a matrix.");
     int* dims = INTEGER(dimsExpr);
-    if ((size_t) dims[0] != data.numObservations || (size_t) dims[1] != control.numTrees) error("Dimensions of state@fit.tree do not match object.");
+    if (static_cast<size_t>(dims[0]) != data.numObservations || static_cast<size_t>(dims[1]) != control.numTrees) error("Dimensions of state@fit.tree do not match object.");
     std::memcpy(REAL(slotExpr), state.treeFits, data.numObservations * control.numTrees * sizeof(double));
     
     slotExpr = GET_ATTR(stateExpr, install("fit.total"));
-    if ((size_t) GET_LENGTH(slotExpr) != data.numObservations) error("Length of state@fit.total does not match object.");
+    if (static_cast<size_t>(GET_LENGTH(slotExpr)) != data.numObservations) error("Length of state@fit.total does not match object.");
     std::memcpy(REAL(slotExpr), state.totalFits, data.numObservations * sizeof(double));
     
     if (data.numTestObservations != 0) {
       slotExpr = GET_ATTR(stateExpr, install("fit.test"));
-      if ((size_t) GET_LENGTH(slotExpr) != data.numTestObservations) error("Length of state@fit.test does not match object.");
+      if (static_cast<size_t>(GET_LENGTH(slotExpr)) != data.numTestObservations) error("Length of state@fit.test does not match object.");
       std::memcpy(REAL(slotExpr), state.totalTestFits, data.numTestObservations * sizeof(double));
     }
     
@@ -861,11 +1039,11 @@ namespace {
     REAL(slotExpr)[0] = state.runningTime;
     
     slotExpr = GET_ATTR(stateExpr, install("trees"));
-    if ((size_t) GET_LENGTH(slotExpr) != control.numTrees) error("Length of state@trees does not match object.");
+    if (static_cast<size_t>(GET_LENGTH(slotExpr)) != control.numTrees) error("Length of state@trees does not match object.");
     
     const char** treeStrings = const_cast<const char**>(state.createTreeStrings(fit));
     for (size_t i = 0; i < control.numTrees; ++i) {
-      SET_STRING_ELT(slotExpr, (int) i, CREATE_STRING_VECTOR(treeStrings[i]));
+      SET_STRING_ELT(slotExpr, static_cast<int>(i), CREATE_STRING_VECTOR(treeStrings[i]));
       delete [] treeStrings[i];
     }
     delete [] treeStrings;
@@ -877,14 +1055,14 @@ namespace {
     const Data& data(fit.data);
     
     SEXP slotExpr = GET_ATTR(stateExpr, install("fit.tree"));
-    std::memcpy(state.treeFits, (const double*) REAL(slotExpr), data.numObservations * control.numTrees * sizeof(double));
+    std::memcpy(state.treeFits, const_cast<const double*>(REAL(slotExpr)), data.numObservations * control.numTrees * sizeof(double));
     
     slotExpr = GET_ATTR(stateExpr, install("fit.total"));
-    std::memcpy(state.totalFits, (const double*) REAL(slotExpr), data.numObservations * sizeof(double));
+    std::memcpy(state.totalFits, const_cast<const double*>(REAL(slotExpr)), data.numObservations * sizeof(double));
     
     if (data.numTestObservations != 0) {
       slotExpr = GET_ATTR(stateExpr, install("fit.test"));
-      std::memcpy(state.totalTestFits, (const double*) REAL(slotExpr), data.numTestObservations * sizeof(double));
+      std::memcpy(state.totalTestFits, const_cast<const double*>(REAL(slotExpr)), data.numTestObservations * sizeof(double));
     }
     
     slotExpr = GET_ATTR(stateExpr, install("sigma"));
@@ -896,7 +1074,7 @@ namespace {
     slotExpr = GET_ATTR(stateExpr, install("trees"));
     const char** treeStrings = ext_stackAllocate(control.numTrees, const char*);
     for (size_t i = 0; i < control.numTrees; ++i) {
-      treeStrings[i] = CHAR(STRING_ELT(slotExpr, (int) i));
+      treeStrings[i] = CHAR(STRING_ELT(slotExpr, static_cast<int>(i)));
     }
     state.recreateTreesFromStrings(fit, treeStrings);
     
@@ -905,10 +1083,12 @@ namespace {
   
   void deleteFit(BARTFit* fit) {
 #ifdef THREAD_SAFE_UNLOAD
-    Rprintf("deleting %p\n", fit);
+    Rprintf("deleting   %p\n", fit);
 #endif
     if (fit == NULL) return;
     
+    ext_rng_destroy(fit->control.rng);
+    
     delete fit->model.treePrior;
     delete fit->model.muPrior;
     delete fit->model.sigmaSqPrior;
