@@ -8,9 +8,9 @@
 
 #include <algorithm>
 
-#include <external/stats.h>
-#include <external/io.h>
 #include <external/alloca.h>
+#include <external/io.h>
+#include <external/random.h>
 
 #include <dbarts/bartFit.hpp>
 #include <dbarts/model.hpp>
@@ -79,7 +79,7 @@ namespace dbarts {
     
     // randomly choose a notBottom node = nodeToChange
     //u=ran1(&idum);
-    size_t nodeIndex = (size_t) ext_simulateIntegerUniformInRange(0, (int64_t) numNotBottomNodes);
+    size_t nodeIndex = ext_rng_simulateUnsignedIntegerUniformInRange(fit.control.rng, 0, numNotBottomNodes);
     Node& nodeToChange(*notBottomNodes[nodeIndex]);
     
     //given the node, choose a new variable for the new rule
@@ -101,8 +101,8 @@ namespace dbarts {
       if (numGoodRules > 0) {
         // draw the rule from list of good ones
         //u=ran1(&idum);
-        uint32_t goodCategoryNumber = (uint32_t) ext_simulateUnsignedIntegerUniformInRange(0, numGoodRules);
-        uint32_t categoryCombinationNumber = (uint32_t) findIndexOfIthPositiveValue(categoryCombinationsAreGood, numCategoryCombinations, goodCategoryNumber);
+        uint32_t goodCategoryNumber = static_cast<uint32_t>(ext_rng_simulateUnsignedIntegerUniformInRange(fit.control.rng, 0, numGoodRules));
+        uint32_t categoryCombinationNumber = static_cast<uint32_t>(findIndexOfIthPositiveValue(categoryCombinationsAreGood, numCategoryCombinations, goodCategoryNumber));
         
         
         //get logpri and logL from current tree (X)
@@ -120,13 +120,13 @@ namespace dbarts {
         nodeToChange.p.rule.variableIndex = newVariableIndex;
         nodeToChange.p.rule.categoryDirections = 0u;
         for (size_t j = 0; j < firstGoodCategory; ++j) {
-          if (sel[j] == true) nodeToChange.p.rule.setCategoryGoesRight((uint32_t) j);
-          else nodeToChange.p.rule.setCategoryGoesLeft((uint32_t) j);
+          if (sel[j] == true) nodeToChange.p.rule.setCategoryGoesRight(static_cast<uint32_t>(j));
+          else nodeToChange.p.rule.setCategoryGoesLeft(static_cast<uint32_t>(j));
         }
         nodeToChange.p.rule.setCategoryGoesRight(firstGoodCategory);
         for (size_t j = firstGoodCategory + 1; j < numCategories; ++j) {
-          if (sel[j - 1] == true) nodeToChange.p.rule.setCategoryGoesRight((uint32_t)  j);
-          else nodeToChange.p.rule.setCategoryGoesLeft((uint32_t) j);
+          if (sel[j - 1] == true) nodeToChange.p.rule.setCategoryGoesRight(static_cast<uint32_t>(j));
+          else nodeToChange.p.rule.setCategoryGoesLeft(static_cast<uint32_t>(j));
         }
         
         // fix data at nodes below nodeToChange given new rule
@@ -144,7 +144,7 @@ namespace dbarts {
         alpha = std::exp(YLogPi + YLogL - XLogPi - XLogL);
         alpha = (alpha > 1.0 ? 1.0 : alpha);
         
-        if (ext_simulateBernoulli(alpha) == 1) {
+        if (ext_rng_simulateBernoulli(fit.control.rng, alpha) == 1) {
           oldState.destroy();
           
           *stepTaken = true;
@@ -172,7 +172,7 @@ namespace dbarts {
       
       // if there are any rules
       if (numSplitVariables > 0) {
-        int32_t newRuleIndex = (int32_t) ext_simulateIntegerUniformInRange(left, right + 1);
+        int32_t newRuleIndex = static_cast<int32_t>(ext_rng_simulateIntegerUniformInRange(fit.control.rng, left, right + 1));
         
         //get logpri and logL from current tree (X)
         XLogPi = fit.model.treePrior->computeTreeLogProbability(fit, tree);
@@ -198,7 +198,7 @@ namespace dbarts {
         alpha = std::exp(YLogPi + YLogL - XLogPi - XLogL);
         alpha = (alpha > 1.0 ? 1.0 : alpha);
         
-        if (ext_simulateBernoulli(alpha) == 1) {	
+        if (ext_rng_simulateBernoulli(fit.control.rng, alpha) == 1) {	
           oldState.destroy();
           *stepTaken = true;
         } else {
@@ -227,7 +227,7 @@ namespace dbarts {
     } else {
       if (curr->p.rule.variableIndex == variableIndex) {
 //        if (curr->rule.categoryDirections[categoryIndex] == BART_CAT_RIGHT) {
-        if (curr->p.rule.categoryGoesRight((uint32_t) categoryIndex)) {
+        if (curr->p.rule.categoryGoesRight(static_cast<uint32_t>(categoryIndex))) {
           findReachableBottomNodesForCategory(curr->getRightChild(), variableIndex, categoryIndex, bottomVector, nodesAreReachable);
         } else {
           findReachableBottomNodesForCategory(curr->getLeftChild(), variableIndex, categoryIndex, bottomVector, nodesAreReachable);
@@ -244,7 +244,7 @@ namespace dbarts {
   {
     int32_t leftIndex, rightIndex; 
     leftIndex = 0; // left value if you top out
-    rightIndex = (int32_t) fit.scratch.numCutsPerVariable[variableIndex] - 1; // right value if you top out
+    rightIndex = static_cast<int32_t>(fit.scratch.numCutsPerVariable[variableIndex]) - 1; // right value if you top out
     
     int32_t leftMin, leftMax, rightMin, rightMax;
     
@@ -292,7 +292,7 @@ namespace dbarts {
     bool* categoriesGoRight = ext_stackAllocate(numCategories, bool);
     setCategoryReachability(fit, node, variableIndex, categoriesGoRight);
     
-    *firstGoodCategory = (uint32_t) getIndexOfFirstTrueValue(categoriesGoRight, numCategories);
+    *firstGoodCategory = static_cast<uint32_t>(getIndexOfFirstTrueValue(categoriesGoRight, numCategories));
     
     if (*firstGoodCategory == numCategories) ext_printf("error in findGoodCategoricalRule: no available categories\n");
     
@@ -315,7 +315,7 @@ namespace dbarts {
     
     // enumerate through all possible category branches and test them for validity
     for (uint64_t i = 0; i < numCategoryCombinations; ++i) {
-      setBinaryRepresentation(numCategories - 1, (uint32_t) i, sel1);
+      setBinaryRepresentation(numCategories - 1, static_cast<uint32_t>(i), sel1);
       
       for (size_t j = 0; j < *firstGoodCategory; ++j) sel[j] = sel1[j];
       for (size_t j = *firstGoodCategory + 1; j < numCategories; ++j) sel[j] = sel1[j - 1];
@@ -384,7 +384,7 @@ namespace {
     state.observationIndicesPtrs[nodeIndex] = node.observationIndices;
     state.numObservations[nodeIndex] = node.numObservations;
     state.observationIndices[nodeIndex] = new size_t[node.numObservations];
-    std::memcpy(state.observationIndices[nodeIndex], (const size_t*) node.observationIndices, node.numObservations * sizeof(size_t));
+    std::memcpy(state.observationIndices[nodeIndex], const_cast<const size_t*>(node.observationIndices), node.numObservations * sizeof(size_t));
     
     ++nodeIndex;
     
@@ -403,7 +403,7 @@ namespace {
 
     node.observationIndices = state.observationIndicesPtrs[nodeIndex];
     node.numObservations = state.numObservations[nodeIndex];
-    std::memcpy(node.observationIndices, (const size_t*) state.observationIndices[nodeIndex], state.numObservations[nodeIndex] * sizeof(size_t));
+    std::memcpy(node.observationIndices, const_cast<const size_t*>(state.observationIndices[nodeIndex]), state.numObservations[nodeIndex] * sizeof(size_t));
     
     ++nodeIndex;
     
